#!/usr/bin/env python
cimport cython
cimport numpy
import numpy
import pique

DTYPE = numpy.float64
ctypedef numpy.float64_t DTYPE_t

@cython.boundscheck(False)
def peakdet( v,         \
             delta,     \
             x = None):
    """
    Converted from MATLAB script at http://billauer.co.il/peakdet.html
    
    Currently returns two lists of tuples, but maybe arrays would be better
    
    function [maxtab, mintab]=peakdet(v, delta, x)
    %PEAKDET Detect peaks in a vector
    %        [MAXTAB, MINTAB] = PEAKDET(V, DELTA) finds the local
    %        maxima and minima ("peaks") in the vector V.
    %        MAXTAB and MINTAB consists of two columns. Column 1
    %        contains indices in V, and column 2 the found values.
    %      
    %        With [MAXTAB, MINTAB] = PEAKDET(V, DELTA, X) the indices
    %        in MAXTAB and MINTAB are replaced with the corresponding
    %        X-values.
    %
    %        A point is considered a maximum peak if it has the maximal
    %        value, and was preceded (to the left) by a value lower by
    %        DELTA.
    
    % Eli Billauer, 3.4.05 (Explicitly not copyrighted).
    % This function is released to the public domain; Any use is allowed.
    
    """
    v = numpy.asarray( v, dtype=DTYPE )

    maxtab = []
    mintab = []
    
    cdef DTYPE_t mx = -numpy.inf
    cdef DTYPE_t mn =  numpy.inf
    
    cdef DTYPE_t mxpos = numpy.nan
    cdef DTYPE_t mnpos = numpy.nan
    
    cdef DTYPE_t this = 0
    
    if x is None:
        x = numpy.arange(len(v))
    else :    
        x = numpy.asarray( x, dtype=DTYPE )
    
    if len(v) != len(x):
        raise pique.PiqueException( 'Input vectors v and x must have same length')
    
    if not numpy.isscalar(delta):
        raise pique.PiqueException( 'Input argument delta must be a scalar')
    
    if delta <= 0:
        raise pique.PiqueException( 'Input argument delta must be positive')
    
    lookformax = True
    
    for i in numpy.arange(len(v)):
        this = v[i]
        if this > mx:
            mx = this
            mxpos = x[i]
        if this < mn:
            mn = this
            mnpos = x[i]
        
        if lookformax:
            if this < mx-delta:
                maxtab.append((mxpos, mx))
                mn = this
                mnpos = x[i]
                lookformax = False
        else:
            if this > mn+delta:
                mintab.append((mnpos, mn))
                mx = this
                mxpos = x[i]
                lookformax = True

    return maxtab, mintab


@cython.boundscheck(False)
def region( v,      \
            x,      \
            delta,  \
            radius=1000 ) :
    """
    Calculate the points around a peak which fall below delta, where
    delta is the fraction of the maximum height of the peak.
        
        v       : the array contianing the data
        x       : the array coordiate of the peak
        delta   : cutoff ratio
        radius  : terminate serach beyond this radius
        
    """
    
    cdef DTYPE_t x0 = -numpy.inf
    cdef DTYPE_t x1 =  numpy.inf
    
    for x0 in numpy.arange(radius) :
        if v[x-x0] <= v[x] * delta :
            break

    for x1 in numpy.arange(radius) :
        if v[x+x1] <= v[x] * delta :
            break
    
    return x0, x1




"""
nose tests for pique.
"""
import pique.data
import numpy

def test_init_PiqueData() :
    
    # does the data load?
    D = pique.data.PiqueData( 'data/IP.bam', 'data/WCE.bam', 'data/map.gff', name='test' )
    
    # did we get the expected number of contigs?
    assert len( D.data.keys() ) == 3
    
    # are they the ones we were looking for?
    assert D.data.keys().__contains__( 'Chromosome' )
    assert D.data.keys().__contains__( 'PNRC100' )
    assert D.data.keys().__contains__( 'PNRC200' )
    
    # is the track data present?
    for contig in D.data.keys() :
        assert type(D.data[contig]['IP']['forward']) is numpy.ndarray
        assert type(D.data[contig]['IP']['reverse']) is numpy.ndarray
        assert type(D.data[contig]['BG']['forward']) is numpy.ndarray
        assert type(D.data[contig]['BG']['reverse']) is numpy.ndarray

def test_import_cython_stuff() :
    import pique.peak



#!/usr/bin/env python
"""
Pique runtime.
"""
import pique
import sys
import cPickle
import pylab
import numpy

def makemap( name, bamfile, window, stride, highest, lowest, bins ) :
    """
    This function drives the genome map making workflow.
    """
    logfile = name + '.mapmaker.log'
    mapfile = name + '.map.gff'
    
    pique.msg( logfile, 'starting mapmaker for project : ' + name )
    
    pique.msg( logfile, '  -> BAM file    : ' + bamfile      )
    pique.msg( logfile, '  -> map file    : ' + mapfile      )
    pique.msg( logfile, '  -> window      : ' + str(window)  )
    pique.msg( logfile, '  -> stride      : ' + str(stride)  )
    pique.msg( logfile, '  -> bins        : ' + str(bins)    )
    pique.msg( logfile, '  -> highest bin : ' + str(highest) )
    pique.msg( logfile, '  -> lowest bin  : ' + str(lowest)  )
    
    pique.msg( logfile, 'loading data...' )
    
    data = pique.fileIO.loadBAM( bamfile )
    
    pique.msg( logfile, '  found contigs :' )
    
    for contig in data.keys() :
        pique.msg( logfile, '    ' + contig )
        pique.msg( logfile, '      ' + str(len(data[contig]['forward'])) )
    
    pique.msg( logfile, 'making spectral histograms...' )
    
    sh = {}
    for contig in data.keys() :
        pique.msg( logfile, '  :: making sectral histogram for contig ' + contig )
        d = numpy.array( data[contig]['forward'] + data[contig]['reverse'], dtype = int )
        sh[contig] = pique.mapmaker.hist(   d,
                                            lowest,
                                            highest,
                                            bins,
                                            window,
                                            stride     )
    
    # save images of spectral histograms
    pique.msg( logfile, 'saving images of spectral histograms...' )
    
    for contig in sh.keys() :
        pylab.cla() # clean up crumbs from last plot
        pylab.clf() # clean up crumbs from last plot
        pique.msg( logfile, '  :: saving image for contig ' + contig )
        pylab.contourf( sh[contig], bins )
        pylab.title( name + ' : ' + contig )
        imgname = name + '_' + contig + '.png'
        pylab.savefig( imgname, format='png' )
    

def detect( name, ipfile, bgfile, mapfile, alpha, l_thresh, pickle_file ) :
    """
    This function drives the peak detection workflow.
    """
    # set logfile
    logfile = name + '.log'
    
    pique.msg( logfile, 'starting run for project : ' + name )
    
    # log inputs
    pique.msg( logfile, '  -> IP file  : ' + ipfile   )
    pique.msg( logfile, '  -> BG file  : ' + bgfile   )
    pique.msg( logfile, '  -> map file : ' + mapfile  )
    pique.msg( logfile, '  -> alpha    : ' + str(alpha)    )
    pique.msg( logfile, '  -> l_thresh : ' + str(l_thresh) )
    
    # load the data
    pique.msg( logfile, 'loading data...' )
    D = pique.data.PiqueData( ipfile, bgfile, mapfile, name=name )
    
    pique.msg( logfile, '  found contigs :' )
    for contig in D.data.keys() :
        pique.msg( logfile, '    ' + contig )
        pique.msg( logfile, '      length : ' + str(D.data[contig]['length']) )
        for r in D.data[contig]['regions'] :
            start = str( r['start'] )
            stop  = str( r['stop']  )
            pique.msg( logfile, '      analysis region : ' + start + ':' + stop )
        for m in D.data[contig]['masks'] :
            start = str( m['start'] )
            stop  = str( m['stop']  )
            pique.msg( logfile, '      masking region  : ' + start + ':' + stop )
    
    # start analysis workbench
    pique.msg( logfile, 'creating analysis workbench...' )
    PA = pique.analysis.PiqueAnalysis( D )
    
    # run filters
    pique.msg( logfile, 'running filters...' )
    
    for ar_name in PA.data.keys() :
        pique.msg( logfile, '  :: applying filters to analysis region ' + ar_name )
        PA.apply_filter( ar_name, alpha, l_thresh )
    
    # find peaks
    pique.msg( logfile, 'finding peaks...' )
    for ar_name in PA.data.keys() :
        PA.find_peaks(ar_name)
        pique.msg( logfile, '  peaks ' + ar_name + ' : ' + str(len(PA.data[ar_name]['peaks'])) )
        pique.msg( logfile, '     noise threshold  : ' + str(PA.data[ar_name]['N_thresh']) )
        pique.msg( logfile, '     filter threshold : ' + str(PA.data[ar_name]['n_thresh']) )
        pique.msg( logfile, '     normalizations   : ' + ', '.join( map(str, PA.data[ar_name]['norms']) ) )
    
    # if a pickle file was requested, write it
    pique.msg( logfile, 'pickling analysis workbench...' )
    if pickle_file :
        cPickle.dump( PA, open( name + '.pickle', 'w' ) )
    
    # write output files
    pique.msg( logfile, 'writing output files...' )
    pique.fileIO.writepeaksGFF(  name + '.gff',      PA.data )
    pique.fileIO.writebookmarks( name + '.bookmark', PA.data, name=name )
    pique.fileIO.writeQP(        name + '.qp',       PA.data )
    pique.fileIO.writepeakTSV(   name + '.peak.tsv', PA.data )
    pique.fileIO.writetrack(     name + '.IP.track', D.data  )
    pique.fileIO.writetrack(     name + '.BG.track', D.data, track='BG' )

    # done!
    pique.msg( logfile, 'run completed.' )





#!/usr/bin/env python
"""
Pique analysis module.
"""
import processing
import pique
import numpy
import scipy

class PiqueAnalysis :
    """
    Pique analysis workbench object.
    """
    
    def __init__( self, PD ) :
        """
        Workbench initialization requires a populated PiqueData
        object. Otherwise, there's nothing to do!
        """
        self.PD   = PD
        self.name = PD.name
        self.data = {}
        
        # populate local data container
        
        for contig in self.PD.data.keys() :
            
            ip_f = self.PD.data[contig]['IP']['forward']
            ip_r = self.PD.data[contig]['IP']['reverse']
            bg_f = self.PD.data[contig]['BG']['forward']
            bg_r = self.PD.data[contig]['BG']['reverse']
            
            # apply masks
            for mask in self.PD.data[contig]['masks'] :
                l = mask['stop'] - mask['start']
                
                ip_f[ mask['start'] : mask['stop'] ] = numpy.zeros(l)
                ip_r[ mask['start'] : mask['stop'] ] = numpy.zeros(l)
                bg_f[ mask['start'] : mask['stop'] ] = numpy.zeros(l)
                bg_r[ mask['start'] : mask['stop'] ] = numpy.zeros(l)
            
            # calculate normalization factors
            norms = []
            for norm in self.PD.data[contig]['norms'] :
                nip_f = ip_f[ norm['start'] : norm['stop'] ]
                nip_r = ip_r[ norm['start'] : norm['stop'] ]
                nbg_f = bg_f[ norm['start'] : norm['stop'] ]
                nbg_r = bg_r[ norm['start'] : norm['stop'] ]
                nip = float(sum( nip_f + nip_r ))
                nbg = float(sum( nbg_f + nbg_r ))
                if not nbg == 0 :
                    n = nip/nbg
                else :
                    n = None
                norms.append( { 'start' : norm['start'],    \
                                'stop'  : norm['stop'],     \
                                'n'     : n               } )           
                
            # create analysis regons
            for r in self.PD.data[contig]['regions'] :
                
                IP = { 'forward' : ip_f[ r['start'] : r['stop'] ],  \
                       'reverse' : ip_r[ r['start'] : r['stop'] ]   } 
                BG = { 'forward' : bg_f[ r['start'] : r['stop'] ],  \
                       'reverse' : bg_r[ r['start'] : r['stop'] ]   } 
                
                ar = { 'contig' : contig, 'IP' : IP, 'BG' : BG, 'region' : r }
                               
                # attach norm regions
                ar['norms'] = []
                ar['n_regions'] = []
                for norm in norms :
                    start = norm['start']
                    stop  = norm['stop']
                    if r['start'] < start and r['stop'] > stop :
                        ar['norms'].append( norm['n'] )
                        n_start = start - r['start']
                        n_stop  = stop  - r['start']
                        ar['n_regions'].append( { 'start' : n_start, 'stop' : n_stop } )
                
                # calcualte (normalized) noise threshold
                bg_all = numpy.concatenate( ( ar['BG']['forward'], ar['BG']['reverse'] ) )
                N = numpy.mean(ar['norms'])
                if not N == 0 :
                    ar['N_thresh'] = self.noise_threshold( bg_all ) / numpy.mean(ar['norms'])
                else :
                    ar['N_thresh'] = self.noise_threshold( bg_all )
                
                ar['peaks'] = []
                
                name = contig + '_' + str( r['start'] ) + ':' + str( r['stop'] )
                
                self.data[name] = ar
                
    def noise_threshold( self, data ) :
        """
        Computes the noise threshold in an analysis region. For now,
        this is the 90th quantile of the data.
        """
        #return stats.scoreatpercentile( data.tolist(), 90 )
        return sorted(data)[ min( len(data)-1, int(len(data)*0.95)) ]
        
    def apply_filter( self, ar_name, alpha, l_thresh, ) :
        """
        Apply the filter set to an analysis region.
        """
        ar = self.data[ar_name]

        fip_f = processing.filterset( ar['IP']['forward'], alpha, l_thresh )
        fip_r = processing.filterset( ar['IP']['reverse'], alpha, l_thresh )
        fbg_f = processing.filterset( ar['BG']['forward'], alpha, l_thresh )
        fbg_r = processing.filterset( ar['BG']['reverse'], alpha, l_thresh )
        
        self.data[ar_name]['ip'] = { 'forward' : fip_f, 'reverse' : fip_r }
        self.data[ar_name]['bg'] = { 'forward' : fbg_f, 'reverse' : fbg_r }
        
        # calculate level offset between filtered IP and BG tracks,
        # and scale the noise threshold accordingly
        n = []
        for norm in ar['n_regions'] :
            start = norm['start']
            stop  = norm['stop']
            ip = numpy.concatenate( (fip_f[start:stop], fip_r[start:stop] ) )
            bg = numpy.concatenate( (fbg_f[start:stop], fbg_r[start:stop] ) )
            nip = float(sum( ip )) / (stop-start)
            nbg = float(sum( bg )) / (stop-start)
            nn  = nip / nbg
            n.append( self.noise_threshold( bg ) / nn )
        self.data[ar_name]['n_thresh'] = numpy.mean(n)

    def find_peaks( self, ar_name ) :
        """
        Find the peaks in an anlysis region and append them to the
        data dictionary.
        """
        length = len( self.data[ar_name]['bg']['forward'] )
        maxcov = max( numpy.concatenate( (                          \
                        self.data[ar_name]['ip']['forward'],        \
                        self.data[ar_name]['ip']['reverse'] ) ) )
        
        # find local maxima in the forward starnd
        peaks_f = pique.peak.peakdet(                               \
                        self.data[ar_name]['ip']['forward'],        \
                        pique.constants.top_delta*maxcov )[0]

        # find local maxima in the reverse starand
        peaks_r = pique.peak.peakdet(                               \
                        self.data[ar_name]['ip']['reverse'],        \
                        pique.constants.top_delta*maxcov )[0]
        
        # build up a peak annotation for maxima in the forward and
        # reverse strand coverage that occur within the peak
        # separation distance
        for ix, iy in peaks_f :
            for jx, jy in peaks_r :
                if jx - ix > 0 and jx - ix < pique.constants.peak_separation :
                    
                    x = int((ix+jx)/2.0)
                    
                    # find the region around the forward strand
                    # maximum
                    fx0,fx1 = pique.peak.region(                    \
                        self.data[ar_name]['ip']['forward'],        \
                        ix,                                         \
                        pique.constants.reg_delta,                  \
                        radius=pique.constants.radius )
                    
                    # find the region around the reverse strand
                    # maximum
                    rx0,rx1 = pique.peak.region(                    \
                        self.data[ar_name]['ip']['reverse'],        \
                        jx,                                         \
                        pique.constants.reg_delta,                  \
                        radius=pique.constants.radius )
                    
                    # the region start and stop coordinates    
                    start = ix - fx0
                    stop  = jx + rx1
                    
                    # compute the enrichment ratio....
                    ip_e = sum( 
                        self.data[ar_name]['IP']['forward'][start:stop]   \
                      + self.data[ar_name]['IP']['reverse'][start:stop]   )
                    
                    bg_e = sum( 
                        self.data[ar_name]['BG']['forward'][start:stop]   \
                      + self.data[ar_name]['BG']['reverse'][start:stop]   )
                    
                    fip_f = self.data[ar_name]['ip']['forward'][start:stop]
                    fip_r = self.data[ar_name]['ip']['reverse'][start:stop]
                    
                    if not float(bg_e) == 0 :
                        er = float(ip_e) / float(bg_e)
                    else :
                        er = None
                    
                    annotations = {                                 \
                        'binds_at'          : x,                    \
                        'enrichment_ratio'  : er
                    }
                    
                    # normalizations
                    for i,norm in enumerate(self.data[ar_name]['norms']) :
                        name = 'norm_' + str(i)
                        annotations[name] = norm
                    
                    peak = {                                        \
                        'start'             : start,                \
                        'stop'              : stop,                 \
                        'annotations'       : annotations
                    }

                    self.data[ar_name]['peaks'].append(peak)
        
    def find_peaks_depricated( self, ar_name ) :
        """
        Find the peaks in an analysis region.
        
        WARNING : THIS FUNCTION IS DEPRICATED.
        """
        fp = processing.findregions( self.data[ar_name]['ip']['forward'],   \
                                     self.data[ar_name]['n_thresh']         )
        rp = processing.findregions( self.data[ar_name]['ip']['reverse'],   \
                                     self.data[ar_name]['n_thresh']         )
        
        # loop over the peaks that meet the overlap criterion and add
        # annotations for enrichment ratio, and putative binding
        # coordinate and normalizations
        for e in processing.overlaps( fp, rp ) :
            ip_e = sum( self.data[ar_name]['IP']['forward'][e['start']:e['stop']]   \
                      + self.data[ar_name]['IP']['reverse'][e['start']:e['stop']]   )
            
            bg_e = sum( self.data[ar_name]['BG']['forward'][e['start']:e['stop']]   \
                      + self.data[ar_name]['BG']['reverse'][e['start']:e['stop']]   )
            
            fip_f = self.data[ar_name]['ip']['forward'][e['start']:e['stop']]
            fip_r = self.data[ar_name]['ip']['reverse'][e['start']:e['stop']]
            
            if not float(bg_e) == 0 :
                er = float(ip_e) / float(bg_e)
            else :
                er = None
            
            # enrichment ratio
            e['annotations']['enrichment_ratio'] = er
            
            # normalizations
            for i,norm in enumerate(self.data[ar_name]['norms']) :
                name = 'norm_' + str(i)
                e['annotations'][name] = norm
            
            # simple estimate of binding coordinate; centerpoint
            # between forward and reverse maxima
            binds_at = self.data[ar_name]['region']['start']    \
                     + e['start']                               \
                     + int( ( fip_f.argmax() + fip_r.argmax() ) / 2.0 )
            
            e['annotations']['binds_at'] = binds_at
            
            # add the annotated peak to the peak list for this
            # analysis region
            self.data[ar_name]['peaks'].append(e)



#!/usr/bin/env python
"""
"""
import sys
import warnings
import data
import analysis
import fileIO
import processing
import mapmaker
import runtime
import peak
import constants

# suppress annoying scypy warnings
warnings.simplefilter("ignore", DeprecationWarning)

class PiqueException( Exception ) :
    pass

def version() :
    return constants.version

def readtrack( filename ) :
    """
    Read a track file, return a numpy array.
    """
    data = []
    for line in open( filename ) :
        if line.__contains__( '\"' ) :
            continue
        data.append( float( line.split()[2] ) )
    if len( data ) == 0 :
        raise PiqueException( 'Empty input file or parsing error : ' + filename )
    return numpy.array( data )

def write_strandless_track( data, file, track_name ) :
    f = open( file, 'w' )
    f.write( 'sequence\tstrand\tposition\tvalue\n' )
    for n,i in enumerate( data ) :
        if i != 0 :
            f.write( track_name + '\t.\t' + str(n) + '\t' + str(i) + '\n' )
    f.close()

def write_track( data_forward, data_reverse, file, track_name ) :
    f = open( file, 'w' )
    f.write( 'sequence\tstrand\tposition\tvalue\n' )
    for n,i in enumerate( data_forward ) :
        if i != 0 :
            f.write( track_name + '\t+\t' + str(n) + '\t' + str(i) + '\n' )
    for n,i in enumerate( data_reverse ) :
        if i != 0 :
            f.write( track_name + '\t-\t' + str(n) + '\t' + str(i) + '\n' )
    f.close()

def read_track( file, track_name ) :
    data_ff = []
    data_rr = []
    for line in open( filename ) :
        if line.__contains__( '\"' ) :
            continue
        if not line.__contains__( track_name ) :
            continue
        strand = line.split()[1]
        if strand == '+' :
            data_ff.append( float( line.split()[3] ) )
        if strand == '-' :
            data_rr.append( float( line.split()[3] ) )
    return {'forward':numpy.array(data_ff), 'reverse':numpy.array(data_rr) }

def readbookmarks( filename ) :
    """
    Read a GGB bookmark file.
    """
    regions = []
    for line in open( filename ) :
        if line.__contains__('>') :
            continue
        if line.__contains__('\tName\tAnnotation') :
            continue
        chromosome,start,stop,strand,name,annotations = line.strip().split('\t')
        start, stop = map( int, (start,stop) )
        annot = {}
        for annotation in annotations.split() :
            if annotation.__contains__(':') :
                key,value = annotation.split(':')
                annot[key] = value
        regions.append( {   'start':start,          \
                            'stop':stop,            \
                            'strand':strand,        \
                            'chromosome':chromosome,\
                            'annotations':annot     } )
    return regions

def msg( file, message ) :
    sys.stderr.write( message + '\n' )
    f = open( file, 'a' )
    f.write( message + '\n' )
    f.close()



import numpy
cimport numpy
cimport cython
from libc.stdlib cimport *

cdef extern from 'math.h':
    double ceil(double x)
    double floor (double x)

DTYPE = numpy.int # data type for arrays
ctypedef numpy.int_t DTYPE_t

@cython.boundscheck(False)
def hist( numpy.ndarray[DTYPE_t, ndim=1] data,  \
        int lowest,                             \
        int highest,                            \
        int bins,                               \
        int window,                             \
        int stride ) :
    """
    A two-dimensional windowed histogram with strides, used here to
    visualize genome-wide read coverage.

        data    : A 1-D Numpy array of integers representing coverage
        lowest  : Lowest coverage in the histogram
        highest : Higest coverage in the histogram
        bins    : Number of bins in the histogram
        window  : Length of the sliding window
        stride  : Strides between windows
    """
    cdef int overlap  = int(ceil( window / float(stride) ))
    cdef int w = len(data)/stride + overlap
    cdef numpy.ndarray r = numpy.zeros( [ bins, w ], dtype=DTYPE )
    cdef int interval = int(ceil( ( highest - lowest ) / float(bins) ))
    cdef int x
    cdef int i
    cdef int j
    cdef int n
    cdef int m
    for x in range(len(data)) :
        i = data[x]
        if i >= lowest and i <= highest :
            j = (i-lowest)/interval
            n = x/stride
            r[ j, n:(n+overlap) ] += 1
    return r



#!/usr/bin/env python

# Pique version
version         = 0.1

# cache directory for storing temporary and downloaded files
cache_dir       = '/tmp/pique'

# Test data, IP BAM file
test_ip_file    = 'http://files.figshare.com/224619/IP.bam'

# Test data, background BAM file
test_bg_file    = 'http://files.figshare.com/224618/WCE.bam'

# Test data, mapping file
test_map_file   = 'http://files.figshare.com/224617/map.gff'

# Permitted separation between peaks found on the forward and reverse
# strand. This should correspond with the difference between the
# fragment length and the read length.
peak_separation = 100

# The fall-off from maximum needed to indentify a local maximum
top_delta       = 0.05

# The fall-off from an identified local maximum used to identify the
# peak region.
reg_delta       = 0.125

# The maximum radius around an identified local maximum that can be
# included in a peak region.
radius          = 500



#!/usr/bin/env python
"""
Data representations
"""
import numpy
import fileIO
import processing

class PiqueDataException( Exception ) :
    pass

class PiqueData :
    """
    Container class for managing Pique data.
    """
    def __init__( self ) :
        self.data = {}
        self.filtered = {}    
        self.name = ''
    
    def __init__( self, IP_file, BG_file, map_file=None, name='' ) :
        self.data       = {}
        self.filtered   = {}
        self.name       = name
        
        # check the map file for errors now before loading the data
        if map_file :
            gff = fileIO.loadGFF( map_file )

            types = []
            err = 'Overlapping analysis regions : '
            rvs = 'Reversed coodinates in analysis region : '
            types.append( ('regions', err, rvs) )
            err = 'Overlapping masking regions : '
            rvs = 'Reversed coodinates in masking region : '  
            types.append( ('masks', err, rvs) )
            err = 'Overlapping normalization regions : '
            rvs = 'Reversed coodinates in normalization region : '
            types.append( ('norms', err, rvs) )

            for t,err,rvs in types :
                for region in gff[t] :
                    contig = region['contig']
                    start  = region['start']
                    stop   = region['stop']
                    if start > stop :
                        coord = str(start) + ':' + str(stop)
                        raise PiqueDataException( rvs + contig + coord )
                    for r in gff[t] :
                        if r['contig'] == contig :
                            if r['start'] > start and r['start'] < stop or  \
                               r['stop']  > start and r['stop']  < stop :
                                first  = str(start)      + ':' + str(stop)
                                second = str(r['start']) + ':' + str(r['stop'])
                                raise PiqueDataException( err + contig + ' ' + first + '::' + second )
           
        # load the track data 
        self.load_data( IP_file, BG_file )
        
        # load genome map file, if provided
        if map_file :
            
            region_contigs = []
            for region in gff['regions'] :
                contig = region['contig']
                if not self.data.has_key( contig ) :
                    raise PiqueDataException( 'Map file contains unknown contig : ' + contig )
                if not region_contigs.__contains__( contig ) :
                    region_contigs.append( contig )
            
            # remove default regions from mapped contigs
            for contig in region_contigs :
                self.del_analysis_region( contig, 0, self.data[contig]['length'] )
        
            for region in gff['regions'] :
                contig = region['contig']
                start  = region['start']
                stop   = region['stop']
                self.add_analysis_region( contig, start, stop )
            
            # add masks
            for mask in gff['masks'] :
                contig = mask['contig']
                start  = mask['start']
                stop   = mask['stop']
                self.add_mask( contig, start, stop )
            
            # add normalization regions
            for norm in gff['norms'] :
                contig = norm['contig']
                start  = norm['start']
                stop   = norm['stop']
                self.add_norm( contig, start, stop )
    
    def add_contig( self,   contig_name,    \
                            IP_forward,     \
                            IP_reverse,     \
                            BG_forward,     \
                            BG_reverse      ) :
        
        l = map(len, [IP_forward, IP_reverse, BG_forward, BG_reverse ] )
        if not all( x == l[0] for x in l ) :
            raise PiqueDataException( 'Track have different lengths.' )
        
        IP = { 'forward' : IP_forward, 'reverse' : IP_reverse }
        BG = { 'forward' : BG_forward, 'reverse' : BG_reverse }
        
        self.data[ contig_name ] = { 'IP' : IP, 'BG' : BG }
        
        # create a default analysis region spanning the whole contig
        length = len( IP_forward )
        region = { 'start' : 0, 'stop' : length }
        self.data[ contig_name ][ 'length' ]  = length
        self.data[ contig_name ][ 'regions' ] = [ region ]
        
        # create an empty mask list
        self.data[ contig_name ][ 'masks' ] = []
        
        # create an empty norm list
        self.data[ contig_name ][ 'norms' ] = []

    def load_data( self, IP_file, BG_file ) :
        """
        Digest and load data from BAM files containing alignments of
        IP and background reads.
        
        BAM files must be prepared in such a way that the contig names
        are identical, and each IP track must be identical in length
        to its corresponding background contig.
        
        If the file names are valid URLs, attempt to download the data
        to the cache directory specified in pique.constants.cache_dir.
        """
        
        IP_tracks = fileIO.loadBAM( IP_file )
        BG_tracks = fileIO.loadBAM( BG_file )
        
        IP_contigs = IP_tracks.keys()
        BG_contigs = BG_tracks.keys()
        
        IP_contigs.sort()
        BG_contigs.sort()
        
        if not len(IP_contigs) == len(BG_contigs) :
            raise PiqueDataException( 'BG and IP have different number of contigs.' )
        
        if not all( map( lambda x : x[0] == x[1], zip(IP_contigs,BG_contigs) ) ) :
            raise PiqueDataException( 'BG and IP contig names do not match.',   \
                                    { 'IP' : IP_contigs, 'BG' : BG_contigs } )
        
        for contig in IP_contigs :
            IP_forward = IP_tracks[contig]['forward']
            IP_reverse = IP_tracks[contig]['reverse']
            BG_forward = BG_tracks[contig]['forward']
            BG_reverse = BG_tracks[contig]['reverse']
            
            self.add_contig( contig,    IP_forward, \
                                        IP_reverse, \
                                        BG_forward, \
                                        BG_reverse  )
    
    def del_analysis_region( self, contig, start, stop ) :
        """
        Remove an alaysis region from a contig.
        """
        region = { 'start' : start, 'stop' : stop }
        
        if not self.data[contig]['regions'].__contains__( region ) :
            raise PiqueDataException( 'Analysis region does not exist.', region )
        else :
            self.data[contig]['regions'].remove( region )
        
    def add_analysis_region( self, contig, start, stop ) :
        """
        Add an analysis region to a contig. Overlapping regions are
        not allowed.
        """
        
        self.data[contig]['regions'].append( { 'start' : int(start), 'stop' : int(stop) } )
        
    def add_mask( self, contig, start, stop ) :
        """
        Add a mask to a contig. Overlapping masks are not allowed.
        """
        
        self.data[contig]['masks'].append( { 'start' : int(start), 'stop' : int(stop) } )

    def add_norm( self, contig, start, stop ) :
        """
        Add a mask to a contig. Overlapping masks are not allowed.
        """
        
        self.data[contig]['norms'].append( { 'start' : int(start), 'stop' : int(stop) } )




#!/usr/bin/env python
"""
Generate coverage tracks from a BAM file.
"""
import pysam
import numpy
import sys
import urllib
import constants
import os

GFFkeys = ['contig','source','feature','start','stop','score','strand','frame','group']

def downloadURL( url, path ) :
    """
    Download a file.
    
    Code adapted from 
    http://stackoverflow.com/questions/22676/how-do-i-download-a-file-over-http-using-python
    """
    
    # create cache_dir if it doesn't exist already
    if not os.path.isdir( path ) :
        os.mkdir( path )
    
    file = url.split('/')[-1]
    u = urllib.urlopen(url)
    file_path = os.path.join( path, file )

    meta = u.info()
    file_size = int(meta.getheaders("Content-Length")[0])

    # if we've already downloaded the file, don't download it again
    if os.path.isfile( file_path ) :
        if os.path.getsize( path ) == file_size :
            return file_path
    
    f = open( file_path, 'wb')
    
    print "Downloading: %s Bytes: %s" % (file, file_size)
    
    file_size_dl = 0
    block_sz = 8192
    while True:
        
        buffer = u.read(block_sz)
        
        # if the buffer is embty, time to stop
        if not buffer:
            break
        
        file_size_dl += len(buffer)
        f.write(buffer)
        status = r"%10d  [%3.2f%%]" % (file_size_dl, file_size_dl * 100. / file_size)
        status = status + chr(8)*(len(status)+1)
        print status,
        
    f.close()
    
    return file_path

def loadBAM( file ) :
    """
    Read each contig in a BAM file and return a dictionary of tracks.
    Each track contains scalar length, and two vectors representing
    the forward and everse coverage across the contig.
    """

    # if the file path is a URL, download it to the cache directory
    # specified in pique.constants.cache_dir
    if file.split(':')[0].lower() == 'http' :
        file = downloadURL( file, constants.cache_dir )
    
    # index the BAM file
    pysam.index( file )
    
    # open the BAM file
    samfile = pysam.Samfile( file, 'rb' )
    tracks = {}
    
    # loop over the congigs and build forward and reverse coverage
    # tracks for them
    for n,contig in enumerate(samfile.references) :
        
        length = samfile.lengths[n]
        tracks[contig] = {  'length'  : length,                 \
                            'forward' : numpy.zeros(length),    \
                            'reverse' : numpy.zeros(length), }
        
        for pileupcolumn in samfile.pileup( contig, 0, length) :
            
            fn,rn = 0,0
            
            for pileupread in pileupcolumn.pileups :
                
                if pileupread.alignment.is_reverse :
                    rn = rn + 1
                else :
                    fn = fn + 1
                
            tracks[contig]['forward'][pileupcolumn.pos] = fn
            tracks[contig]['reverse'][pileupcolumn.pos] = rn
            
    samfile.close()
    
    return tracks

def loadGFF( file ) :
    """
    Parse a GFF file and return the analysis region and mask features.
    """

    # if the file path is a URL, download it to the cache directory
    # specified in pique.constants.cache_dir
    if file.split(':')[0].lower() == 'http' :
        file = downloadURL( file, constants.cache_dir )

    regions = []
    masks   = []
    norms   = []
    for line in open( file ) :
        g = dict(zip( GFFkeys, line.strip().split('\t') ))
        if g['feature'] == 'analysis_region' :
            region = {  'contig' : g['contig']      ,   \
                        'start'  : int( g['start'] ),   \
                        'stop'   : int( g['stop']  )    }
            regions.append(region)
            
        if g['feature'] == 'mask' :
            mask =  {   'contig' : g['contig']      ,   \
                        'start'  : int( g['start'] ),   \
                        'stop'   : int( g['stop']  )    }
            masks.append(mask)
        
        if g['feature'] == 'norm_region' :
            norm =  {   'contig' : g['contig']      ,   \
                        'start'  : int( g['start'] ),   \
                        'stop'   : int( g['stop']  )    }
            norms.append(norm)

    return { 'regions' : regions, 'masks' : masks, 'norms' : norms }

def writepeaksGFF( file, data ) :
    """
    Write peak list as a GFF file from a PiqueAnalysis.data object.
    """
    f = open( file, 'w' )
    for ar_name in data.keys() :
        ar      = data[ar_name]
        contig  = ar['contig']
        source  = 'Pique-1.0'
        feature = 'peak'
        strand  = '.'
        frame   = '.'
        group   = ''
        rstart  = ar['region']['start']
        for e in ar['peaks'] :
            start = str( rstart + e['start'] )
            stop  = str( rstart + e['stop']  )
            er    = str( e['annotations']['enrichment_ratio'] )
            s = '\t'.join( [contig,source,feature,start,stop,er,strand,frame,group ] )
            f.write( s + '\n' )
    f.close()

def writetrack( file, data, track='IP' ) :
    """
    Write a Gaggle Genome Browser compatible track file from a
    PiqueData.data object.
    """
    f = open( file, 'w' )
    f.write( 'sequence\tstrand\tposition\tvalue\n' )
    for contig in data.keys() :
        for n,i in enumerate( data[contig][track]['forward'] ) :
            if i != 0 :
                f.write( contig + '\t+\t' + str(n) + '\t' + str(i) + '\n' )
        for n,i in enumerate( data[contig][track]['reverse'] ) :
            if i != 0 :
                f.write( contig + '\t-\t' + str(n) + '\t' + str(i) + '\n' )
    f.close()

def writeQP( file, data ) :
    """
    Write GGB quantitative positional file of estimated binding
    coodinates.
    """
    f = open( file, 'w' )
    f.write( 'sequence\tstrand\tposition\tvalue\n' )
    for ar_name in data.keys() :
        ar      = data[ar_name]
        contig  = ar['contig']
        for e in ar['peaks'] :
            er    = str( e['annotations']['enrichment_ratio'] )
            bc    = str( e['annotations']['binds_at'] )
            f.write( contig + '\t'  \
                    + '.\t'         \
                    + bc + '\t'     \
                    + er + '\n'     )
    f.close()

def writepeakTSV( file, data ) :
    """
    Write a TSV file containing
        contig
        peak start
        peak stop
        peak binding coordinate
        peak enrichment ratio
        average contig normalization
        standard deviation of contig normalization
        norm_0
        norm_1
        norm_2
        norm_3
        ...
    """
    with open( file, 'w' ) as f :
        f.write( 'contig\tstart\tstop\tbinds_at\tenrichment_ratio\taverage_contig_norm\tstd_contig_norm\n' )
        for ar_name in data.keys() :
            ar      = data[ar_name]
            rstart  = ar['region']['start']
            contig  = ar['contig']
            norms   = ar['norms']
            n_av    = str(numpy.mean( norms ))
            n_std   = str(numpy.std(  norms ))
            for e in ar['peaks'] :
                start = str( rstart + e['start'] )
                stop  = str( rstart + e['stop']  )
                er    = str(e['annotations']['enrichment_ratio'])
                bc    = str(e['annotations']['binds_at'])
                line  = '\t'.join( [ contig, start, stop, bc, er, n_av, n_std ] )
                line  = line + '\t' + '\t'.join( map( str, norms ) ) + '\n'
                f.write( line )

def writebookmarks( file, data, name='' ) :
    """
    Write a GGB bookmark file from an PiqueAnalyis.data object.
    """
    f = open( file, 'w' )
    f.write( '>name: Pique bookmarks for ' + name + '\n' )
    f.write( 'Chromosome\tStart\tEnd\tStrand\tName\tAnnotation\n' )
    for ar_name in data.keys() :
        ar      = data[ar_name]
        contig  = ar['contig']
        source  = 'Pique-1.0'
        feature = 'peak'
        strand  = '.'
        frame   = '.'
        group   = ''
        rstart  = ar['region']['start']
        for e in ar['peaks'] :
            start = str( rstart + e['start'] )
            stop  = str( rstart + e['stop']  )
            er    = str( e['annotations']['enrichment_ratio'] )
            f.write( contig + '\t'      \
                    + start + '\t'      \
                    + stop  + '\t'      \
                    + 'none\tpeak\t'    )
            if e.has_key( 'annotations' ) :
                for key,value in e['annotations'].items() :
                    f.write( str(key) + ':' + str(value) + ' ' )
            f.write( '\n' )
    f.close()





#!/usr/bin/env python
try:
    from setuptools import setup, find_packages
    from setuptools import Extension 
    from Cython.Distutils import build_ext
except ImportError:
    print '(WARNING: importing distutils, not setuptools!)'
    from distutils.core import setup

ext_modules = [ Extension( 'pique.mapmaker', [ 'pique/mapmaker.pyx' ] ),
                Extension( 'pique.peak'    , [ 'pique/peak.pyx'     ] ) ]

setup(name = 'pique',
    version = '0.1',
    description = 'An efficient peak finder for high coverage ChIP-seq experiments.',
    #long_description=read('README'),
    author = 'Russell Neches',
    author_email = 'ryneches@ucdavis.edu',
    url = 'https://github.com/ryneches/pique',
    packages = ['pique', 'pique.tests'],
    package_data = {'pique.tests': ['test_IP_fwd.txt',  \
                                    'test_IP_rev.txt',  \
                                    'test_WCE_fwd.txt', \
                                    'test_WCE_rev.txt'  ]},
    license = 'BSD',
    test_suite = 'nose.collector',
    cmdclass = { 'build_ext' : build_ext },
    ext_modules = ext_modules,
    classifiers = [
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Programming Language :: Cython',
        'Programming Language :: Python :: 2.7',
        'License :: OSI Approved :: BSD License'
    ],
    requires = [ 'numpy', 'scipy', 'cython', 'pysam', 'matplotlib' ],
    scripts = [ 'scripts/pique' ]
    )



#!/usr/bin/env python
"""
Data processing functions for Pique.
"""
import numpy
import scipy.signal
import itertools
import operator

# numpy burbles about silly things.
numpy.seterr( all='ignore' )

def findregions( data, N ) :
    """
    Return all the regions of an array that exceed N.
    """
    pos = filter(lambda(a) : a[1] == abs(a[1]), zip(range(len(data)),data-N) )
    if len(pos) == 0 :
        return []
    pos = numpy.array( zip(*pos)[0] )
    regions = []
    for k,g in itertools.groupby( enumerate(pos), lambda(i,x):i-x ) :
        l = map( operator.itemgetter(1),g )
        regions.append( {'start':l[0],'stop':l[-1] } )
    return regions

def sortbyintegral( data, regions ) :
    """
    Sort regions of an array according to their definite integrals.
    """
    return sorted( regions, key=lambda r: sum(data[r['start']:r['stop']]) )
    
def filterset( data, alpha, l_thresh ) :
    """
    Core data filtering function.
    """
    
    window = scipy.signal.blackman( l_thresh )
    window = window / sum(window)

    data = scipy.signal.wiener( data, mysize=alpha )
    data = scipy.signal.convolve( data, window, mode='same' )
    
    return data

def overlaps( forward_regions, reverse_regions ) :
    """
    Find enriched regions that meet the overlap criterion.
    """
    envelopes = []
    for l in forward_regions :
        for m in reverse_regions :
            if m['start'] < l['stop'] < m['stop'] :
                if l['start'] < m['start'] < l['stop'] :
                    e = {   'forward':l,        \
                            'reverse':m,        \
                            'start':l['start'], \
                            'stop':m['stop'],   \
                            'annotations':{}    }
                    envelopes.append( e )
    return envelopes

